package main

import (
	"context"
	"net"
	"net/http"
	"time"

	address "gitlab.dcloud.tech/mashplace/backend/address/api"
	"gitlab.dcloud.tech/mashplace/backend/address/internal/model/mongo"
	"gitlab.dcloud.tech/mashplace/backend/address/internal/service"
	"gitlab.dcloud.tech/mashplace/backend/address/pkg/idealpostcodes"
	"gitlab.dcloud.tech/mashplace/backend/address/version"
	"gitlab.dcloud.tech/mashplace/backend/common/api/health"
	"gitlab.dcloud.tech/mashplace/backend/common/generic/addr"
	"gitlab.dcloud.tech/mashplace/backend/common/generic/env"
	"gitlab.dcloud.tech/mashplace/backend/common/generic/error/mongoerr"
	"gitlab.dcloud.tech/mashplace/backend/common/generic/grpc"
	logger "gitlab.dcloud.tech/mashplace/backend/common/generic/log"
	"gitlab.dcloud.tech/mashplace/backend/common/generic/metrics"
	"gitlab.dcloud.tech/mashplace/backend/common/generic/tracer"
)

var (
	envMongoURL      = env.ResolveString("MONGO_URL", "mongodb:27017")
	envMongoUsername = env.ResolveString("MONGO_USERNAME", "extras-data")
	envMongoPassword = env.ResolveString("MONGO_PASSWORD", "extras-data")
	envMongoAuthDB   = env.ResolveString("MONGO_AUTH", "admin")
	envMongoDB       = env.ResolveString("MONGO_DATABASE", "extras-data")
	envMongoSSLMode  = env.ResolveBool("MONGO_SSL_MODE", false)
)

var (
	envIdealpostcodesKey = env.ResolveString(
		"IDEAL_POSTCODES_API_KEY",
		idealpostcodes.DefaultAPIKey,
	)
	envIdealpostcodesRequestTimeout = env.ResolveDuration(
		"IDEAL_POSTCODES_TIMEOUT",
		3*time.Second,
	)
	envIdealpostcodesQueryURL = env.ResolveString(
		"IDEAL_POSTCODES_QUERY_URL",
		idealpostcodes.DefaultAPIQueryURL,
	)
	envIdealpostcodesPostcodesURL = env.ResolveString(
		"IDEAL_POSTCODES_POSTCODES_URL",
		idealpostcodes.DefaultAPIPostcodesURL,
	)
	envIdealpostcodesAutocompleteURL = env.ResolveString(
		"IDEAL_POSTCODES_AUTOCOMPLETE_URL",
		idealpostcodes.DefaultAPIAutocompleteURL,
	)
	envIdealpostcodesBaseURL = env.ResolveString(
		"IDEAL_POSTCODES_BASE_URL",
		idealpostcodes.DefaultAPIBaseURL,
	)
)

func main() {
	ctx := context.Background()

	defer tracer.Close()

	repository, err := mongo.NewRepository(ctx, &mongo.Config{
		URL:          envMongoURL,
		Username:     envMongoUsername,
		Password:     envMongoPassword,
		Database:     envMongoDB,
		AuthSource:   envMongoAuthDB,
		MongoSSLMode: envMongoSSLMode,
	})
	if err != nil {
		panic(err)
	}

	err = repository.Migrate(ctx)
	if err != nil {
		panic(err)
	}

	srv, err := service.NewService(ctx, &service.Config{
		Repository: repository,
		IdealPostcodes: idealpostcodes.NewClient(&idealpostcodes.Config{
			QueryURL:        envIdealpostcodesQueryURL,
			PostcodesURL:    envIdealpostcodesPostcodesURL,
			AutocompleteURL: envIdealpostcodesAutocompleteURL,
			BaseURL:         envIdealpostcodesBaseURL,
			Key:             envIdealpostcodesKey,
			Client:          &http.Client{},
			RequestTimeout:  envIdealpostcodesRequestTimeout,
		}),
	})
	if err != nil {
		panic(err)
	}

	listener, err := net.Listen("tcp", addr.Grpc)
	if err != nil {
		panic(err)
	}

	grpcServer, err := grpc.NewServer("address", grpc.ServerOptionErrorTranslator(
		mongoerr.Translate,
	))
	if err != nil {
		panic(err)
	}

	address.RegisterAddressServiceServer(grpcServer, srv)

	logger.Infoln("address started")

	healthServer := version.NewHealthServer()
	health.RegisterHealthServiceServer(grpcServer, healthServer)

	go func() {
		_ = metrics.ConfigureListenAndServe(grpcServer)
	}()

	err = grpcServer.Serve(listener)
	if err != nil {
		logger.WithContext(ctx).Errorln("gRPC error", err)
	}
}
