// Package mongo provides mongo implementation for persistence repository
package mongo

import (
	"context"
	"crypto/tls"
	"fmt"

	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"

	"gitlab.dcloud.tech/mashplace/backend/address/internal/model"
	v1 "gitlab.dcloud.tech/mashplace/backend/address/internal/model/mongo/migrations/v1"
	v2 "gitlab.dcloud.tech/mashplace/backend/address/internal/model/mongo/migrations/v2"
	"gitlab.dcloud.tech/mashplace/backend/common/generic/migrations"
	migrationsMongo "gitlab.dcloud.tech/mashplace/backend/common/generic/migrations/mongo"
)

const (
	collectionPostcodeAddress = "postcode_address"
	collectionQueryAddress    = "query_address"
	collectionUdprnAddress    = "udprn_address"
)

// Config mongo repository config
type Config struct {
	URL          string
	Username     string
	Password     string
	Database     string
	AuthSource   string
	MongoSSLMode bool
}

type repositoryImpl struct {
	client                    *mongo.Client
	database                  *mongo.Database
	collectionPostcodeAddress *mongo.Collection
	collectionQueryAddress    *mongo.Collection
	collectionUdprnAddress    *mongo.Collection
}

func validateConfig(config *Config) error {
	if config == nil {
		return fmt.Errorf("invalid repository config")
	}

	if config.URL == "" {
		return fmt.Errorf("empty url in repository config")
	}

	if config.Username == "" {
		return fmt.Errorf("empty username in repository config")
	}

	if config.Password == "" {
		return fmt.Errorf("empty password in repository config")
	}

	if config.Database == "" {
		return fmt.Errorf("empty database in repository config")
	}

	if config.AuthSource == "" {
		return fmt.Errorf("empty database in repository config")
	}

	return nil
}

// NewRepository returns new repository instance
func NewRepository(ctx context.Context, config *Config) (model.Repository, error) {
	err := validateConfig(config)
	if err != nil {
		return nil, err
	}

	clientOptions := options.Client().ApplyURI(fmt.Sprintf("mongodb+srv://%s", config.URL)).
		SetAuth(options.Credential{
			AuthSource:  config.AuthSource,
			Username:    config.Username,
			Password:    config.Password,
			PasswordSet: true,
		})

	if config.MongoSSLMode {
		clientOptions.SetTLSConfig(&tls.Config{})
	}

	client, err := mongo.NewClient(clientOptions)
	if err != nil {
		return nil, err
	}

	err = client.Connect(ctx)
	if err != nil {
		return nil, err
	}

	db := client.Database(config.Database)

	return &repositoryImpl{
		client:                    client,
		database:                  db,
		collectionPostcodeAddress: db.Collection(collectionPostcodeAddress),
		collectionQueryAddress:    db.Collection(collectionQueryAddress),
		collectionUdprnAddress:    db.Collection(collectionUdprnAddress),
	}, nil
}

func (repo *repositoryImpl) Migrate(ctx context.Context) error {
	err := migrations.MigrateSchema(migrationsMongo.WithInstance(ctx, repo.database),
		&migrationsMongo.Migration{
			Version: "v1",
			Up:      v1.Up,
		},
		&migrationsMongo.Migration{
			Version: "v2",
			Up:      v2.Up,
		},
	)
	if err != nil {
		return err
	}

	return nil
}

func (repo *repositoryImpl) ResultsPostcodeGetByKey(ctx context.Context, key string) (r *model.Results, err error) {
	r = &model.Results{}

	err = repo.collectionPostcodeAddress.FindOne(ctx, bson.M{
		"key": key,
	}).Decode(r)

	return
}

func (repo *repositoryImpl) ResultsPostcodeSave(ctx context.Context, results *model.Results) error {
	_, err := repo.collectionPostcodeAddress.ReplaceOne(
		ctx,
		bson.M{"key": results.Key},
		results,
		options.Replace().SetUpsert(true),
	)

	return err
}

func (repo *repositoryImpl) ResultsQueryGetByKey(ctx context.Context, key string) (r *model.Results, err error) {
	r = &model.Results{}

	err = repo.collectionQueryAddress.FindOne(ctx, bson.M{
		"key": key,
	}).Decode(r)

	return
}

func (repo *repositoryImpl) ResultsQuerySave(ctx context.Context, results *model.Results) error {
	_, err := repo.collectionQueryAddress.ReplaceOne(
		ctx,
		bson.M{"key": results.Key},
		results,
		options.Replace().SetUpsert(true),
	)

	return err
}

func (repo *repositoryImpl) ResultsUdprnGetByKey(ctx context.Context, key string) (r *model.Results, err error) {
	r = &model.Results{}

	err = repo.collectionUdprnAddress.FindOne(ctx, bson.M{
		"key": key,
	}).Decode(r)

	return
}

func (repo *repositoryImpl) ResultsUdprnSave(ctx context.Context, results *model.Results) error {
	_, err := repo.collectionUdprnAddress.ReplaceOne(
		ctx,
		bson.M{"key": results.Key},
		results,
		options.Replace().SetUpsert(true),
	)

	return err
}
