// Package v2 provides migration for mongo repository
package v2

import (
	"context"

	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
)

// Up applies migration
func Up(ctx context.Context, db *mongo.Database, migrationID string) (int, error) {
	postcodes := db.Collection("postcode_address")
	queries := db.Collection("query_address")
	udprns := db.Collection("udprn_address")

	_, _ = postcodes.Indexes().DropAll(ctx)
	_, _ = queries.Indexes().DropAll(ctx)
	_, _ = udprns.Indexes().DropAll(ctx)

	_, err := postcodes.UpdateMany(ctx, bson.M{}, bson.M{
		"$rename": bson.M{
			"postcode": "key",
		},
	})
	if err != nil {
		return -1, err
	}

	_, err = queries.UpdateMany(ctx, bson.M{}, bson.M{
		"$rename": bson.M{
			"query": "key",
		},
	})
	if err != nil {
		return -1, err
	}

	_, err = udprns.UpdateMany(ctx, bson.M{}, bson.M{
		"$rename": bson.M{
			"udprn_url": "key",
		},
	})
	if err != nil {
		return -1, err
	}

	_, err = udprns.UpdateMany(ctx, bson.M{},
		bson.A{
			bson.M{
				"$set": bson.M{
					"addresses": bson.A{"$addresses"},
				},
			},
		},
	)
	if err != nil {
		return -1, err
	}

	aggr := bson.A{
		bson.M{
			"$group": bson.M{
				"_id": "$key",
				"count": bson.M{
					"$sum": 1,
				},
			},
		},
		bson.M{
			"$match": bson.M{
				"count": bson.M{
					"$gt": 1,
				},
			},
		},
	}

	type res struct {
		ID string `bson:"_id"`
	}

	for _, c := range []*mongo.Collection{postcodes, queries, udprns} {
		var ress []res

		var cur *mongo.Cursor

		cur, err = c.Aggregate(ctx, aggr)
		if err != nil {
			return -1, err
		}

		err = cur.All(ctx, &ress)
		if err != nil {
			return -1, err
		}

		for _, v := range ress {
			_, err = c.DeleteMany(ctx, bson.M{"key": v.ID})
			if err != nil {
				return -1, err
			}
		}
	}

	_, err = postcodes.Indexes().CreateOne(ctx, mongo.IndexModel{
		Keys:    bson.D{{Key: "key", Value: 1}},
		Options: options.Index().SetUnique(true),
	})
	if err != nil {
		return -1, err
	}

	_, err = queries.Indexes().CreateOne(ctx, mongo.IndexModel{
		Keys:    bson.D{{Key: "key", Value: 1}},
		Options: options.Index().SetUnique(true),
	})
	if err != nil {
		return -1, err
	}

	_, err = udprns.Indexes().CreateOne(ctx, mongo.IndexModel{
		Keys:    bson.D{{Key: "key", Value: 1}},
		Options: options.Index().SetUnique(true),
	})
	if err != nil {
		return -1, err
	}

	return -1, nil
}
