package service

import (
	"context"
	"fmt"
	"regexp"
	"strings"
	"time"

	address "gitlab.dcloud.tech/mashplace/backend/address/api"
	"gitlab.dcloud.tech/mashplace/backend/address/internal/model"
	"gitlab.dcloud.tech/mashplace/backend/address/pkg/idealpostcodes"
	"gitlab.dcloud.tech/mashplace/backend/common/api/grpcerr"
	"gitlab.dcloud.tech/mashplace/backend/common/generic/log"
)

var removeWhitespace = regexp.MustCompile(`\s+`)

type serviceImpl struct {
	config Config
}

func (srv *serviceImpl) GetAddressesByPostcode(
	ctx context.Context,
	req *address.PostcodeRequest,
) (*address.AddressesResponse, error) {
	postcode := removeWhitespace.ReplaceAllString(req.GetPostcode(), "")

	if strings.Contains(postcode, "/") {
		return nil, fmt.Errorf("%w: postcode contain slash char", grpcerr.InvalidData)
	}

	res := &model.Results{}

	if len(postcode) > 0 {
		var err error

		res, err = srv.config.Repository.ResultsPostcodeGetByKey(ctx, postcode)
		if err != nil {
			var addr []*idealpostcodes.Address

			addr, err = srv.config.IdealPostcodes.AddressesByPostcode(ctx, postcode)
			if err != nil {
				log.WithContext(ctx).WithError(err).Error("idealpostcodes error")
			}

			res = &model.Results{
				UpdatedAt: time.Now(),
				Key:       postcode,
				Addresses: srv.createAddresses(addr),
			}

			err = srv.config.Repository.ResultsPostcodeSave(ctx, res)
			if err != nil {
				log.WithContext(ctx).WithError(err).Error("save error")
			}
		}
	}

	return &address.AddressesResponse{
		Addresses: srv.renderAddresses(res.Addresses),
	}, nil
}

func (srv *serviceImpl) GetAddressesByQuery(
	ctx context.Context,
	req *address.QueryRequest,
) (*address.AddressesResponse, error) {
	query := strings.TrimSpace(req.GetQuery())
	if len([]rune(query)) < 2 {
		return nil, fmt.Errorf("%w: addresses query is shorter than 2 characters", grpcerr.InvalidData)
	}

	res, err := srv.config.Repository.ResultsQueryGetByKey(ctx, query)
	if err != nil {
		var addr []*idealpostcodes.Address

		addr, err = srv.config.IdealPostcodes.AddressesByFreeformQuery(ctx, query)
		if err != nil {
			return nil, err
		}

		res = &model.Results{
			UpdatedAt: time.Now(),
			Key:       query,
			Addresses: srv.createAddresses(addr),
		}

		err = srv.config.Repository.ResultsQuerySave(ctx, res)
		if err != nil {
			return nil, err
		}
	}

	return &address.AddressesResponse{
		Addresses: srv.renderAddresses(res.Addresses),
	}, nil
}

func (srv *serviceImpl) AutocompleteAddressByQuery(
	ctx context.Context,
	req *address.QueryRequest,
) (*address.AutocompleteAddressResponse, error) {
	query := strings.TrimSpace(req.GetQuery())

	if len([]rune(query)) < 2 {
		return &address.AutocompleteAddressResponse{}, nil
	}

	suggestions, err := srv.config.IdealPostcodes.AutocompleteAddressByFreeformQuery(ctx, query)
	if err != nil {
		return nil, err
	}

	return &address.AutocompleteAddressResponse{
		Suggestions: srv.renderSuggestions(suggestions),
	}, nil
}

func (srv *serviceImpl) GetAddressByUdprn(
	ctx context.Context,
	req *address.UdprnRequest,
) (*address.UdprnResponse, error) {
	url := strings.TrimSpace(req.UdprnUrl)

	if !strings.HasPrefix(url, "/v1/udprn/") {
		return nil, grpcerr.InvalidData
	}

	res, err := srv.config.Repository.ResultsUdprnGetByKey(ctx, url)
	if err != nil {
		var addr *idealpostcodes.Address

		addr, err = srv.config.IdealPostcodes.AddressByUdprnURL(ctx, url)
		if err != nil {
			return nil, err
		}

		res = &model.Results{
			UpdatedAt: time.Now(),
			Key:       url,
			Addresses: srv.createAddresses([]*idealpostcodes.Address{addr}),
		}

		err = srv.config.Repository.ResultsUdprnSave(ctx, res)
		if err != nil {
			return nil, err
		}
	}

	if len(res.Addresses) == 0 {
		return &address.UdprnResponse{}, nil
	}

	return &address.UdprnResponse{
		Address: srv.renderAddress(res.Addresses[0]),
	}, nil
}
