// Package idealpostcodes provides ideal postcodes client implementation
package idealpostcodes

import (
	"context"
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"net/http/httputil"
	"net/url"
	"strings"
	"time"

	"gitlab.dcloud.tech/mashplace/backend/common/api/grpcerr"
	"gitlab.dcloud.tech/mashplace/backend/common/generic/log"
)

const (
	// DefaultAPIPostcodesURL contains URL to access Postcode Lookup API
	DefaultAPIPostcodesURL = "https://api.ideal-postcodes.co.uk/v1/postcodes"
	// DefaultAPIQueryURL contains URL to access Address Search API
	DefaultAPIQueryURL = "https://api.ideal-postcodes.co.uk/v1/addresses"
	// DefaultAPIKey contains default secret to access paid API
	DefaultAPIKey = "ak_k0qovrm8jpmZJ0yoeOxYSwILcnDqE"
	// DefaultAPIAutocompleteURL contains URL to provides a JSON interface to extract address suggestions for a complete or
	// partial address query.
	DefaultAPIAutocompleteURL = "https://api.ideal-postcodes.co.uk/v1/autocomplete/addresses"
	// DefaultAPIBaseURL contains home url address without endpoints.
	DefaultAPIBaseURL = "https://api.ideal-postcodes.co.uk"
)

// APIStatusOK is success API response code
const apiStatusOK = 2000

type client struct {
	Config
}

// Config provides configuration for ideal postcodes client
type Config struct {
	QueryURL        string
	PostcodesURL    string
	AutocompleteURL string
	BaseURL         string
	Key             string
	Client          *http.Client
	RequestTimeout  time.Duration
}

// NewClient build Client interface implementation.
func NewClient(config *Config) Client {
	return &client{
		Config: *config,
	}
}

// AddressesByPostcode performs request to IdealPostcodes API and returns addresses
// associated with specified postcode.
//
// On failure resulting error wraps:
//   - idealpostcodes.ErrInvalidArgument, if query contains errors
//   - idealpostcodes.ErrProviderError, if response from external service provider is erroneous
func (c *client) AddressesByPostcode(ctx context.Context, postcode string) ([]*Address, error) {
	if postcode = strings.TrimSpace(postcode); postcode == "" {
		return nil, fmt.Errorf("%w: postcode is empty", grpcerr.InvalidData)
	}

	if strings.Contains(postcode, "/") {
		return nil, fmt.Errorf("%w: postcode contain slash char", grpcerr.InvalidData)
	}

	data, err := c.apiGet(ctx, c.PostcodesURL+"/"+postcode, nil)
	if err != nil {
		return nil, err
	}

	response := struct {
		Result  []*Address `json:"result"`
		Code    float64    `json:"code"`
		Message string     `json:"message"`
	}{}

	if err = json.Unmarshal(data, &response); err != nil {
		return nil, err
	}

	if err = apiError(response.Code, response.Message); err != nil {
		return nil, err
	}

	return response.Result, nil
}

// AddressesByFreeformQuery performs request to IdealPostcodes API and returns addresses
// associated with specified freeform query.
//
// On failure resulting error wraps:
//   - idealpostcodes.ErrInvalidArgument, if query contains errors
//   - idealpostcodes.ErrProviderError, if response from external service provider is erroneous
func (c *client) AddressesByFreeformQuery(ctx context.Context, query string) ([]*Address, error) {
	if query = strings.TrimSpace(query); query == "" {
		return nil, fmt.Errorf("%w: addresses query is empty", grpcerr.InvalidData)
	}

	data, err := c.apiGet(ctx, c.QueryURL, url.Values{"query": {query}})
	if err != nil {
		return nil, err
	}

	response := struct {
		Result struct {
			Hits []*Address `json:"hits"`
		} `json:"result"`
		Code    float64 `json:"code"`
		Message string  `json:"message"`
	}{}

	if err = json.Unmarshal(data, &response); err != nil {
		return nil, err
	}

	if err = apiError(response.Code, response.Message); err != nil {
		return nil, err
	}

	var (
		relevant []*Address
		city     string
	)

	for _, hit := range response.Result.Hits {
		if hit.PostTown == "" {
			continue
		}

		if city == "" {
			city = hit.PostTown
		}

		if strings.EqualFold(city, hit.PostTown) {
			relevant = append(relevant, hit)
		}
	}

	return relevant, nil
}

func (c *client) AutocompleteAddressByFreeformQuery(
	ctx context.Context,
	query string,
) ([]*AutocompleteSuggestion, error) {
	if query = strings.TrimSpace(query); query == "" {
		return nil, fmt.Errorf("%w: addresses query is empty", grpcerr.InvalidData)
	}

	data, err := c.apiGet(ctx, c.AutocompleteURL, url.Values{"query": {query}})
	if err != nil {
		return nil, err
	}

	response := struct {
		Result struct {
			Hits []*struct {
				Suggestion string `json:"suggestion"`
				Urls       struct {
					Udprn string `json:"udprn"`
					Umprn string `json:"umprn"`
				} `json:"urls"`
				Udprn int32 `json:"udprn"`
				Umprn int32 `json:"umprn"`
			} `json:"hits"`
		} `json:"result"`
		Code    float64 `json:"code"`
		Message string  `json:"message"`
	}{}

	if err = json.Unmarshal(data, &response); err != nil {
		return nil, err
	}

	if err = apiError(response.Code, response.Message); err != nil {
		return nil, err
	}

	var suggestions []*AutocompleteSuggestion

	for _, s := range response.Result.Hits {
		suggestions = append(suggestions, &AutocompleteSuggestion{
			Suggestion: s.Suggestion,
			UdprnURL:   s.Urls.Udprn,
		})
	}

	return suggestions, nil
}

func (c *client) AddressByUdprnURL(ctx context.Context, udprnURL string) (*Address, error) {
	if udprnURL = strings.TrimSpace(udprnURL); udprnURL == "" {
		return nil, fmt.Errorf("%w: udprn url is empty", grpcerr.InvalidData)
	}

	data, err := c.apiGet(ctx, c.BaseURL+udprnURL, url.Values{})
	if err != nil {
		return nil, err
	}

	response := struct {
		Result  *Address `json:"result"`
		Code    float64  `json:"code"`
		Message string   `json:"message"`
	}{}

	if err = json.Unmarshal(data, &response); err != nil {
		return nil, err
	}

	if err = apiError(response.Code, response.Message); err != nil {
		return nil, err
	}

	return response.Result, nil
}

// apiGet performs HTTP Get request with specified URL.
// Method builds query string by replacing original query values with given params
// and automatically adds API key to the request.
func (c *client) apiGet(ctx context.Context, url string, params url.Values) ([]byte, error) {
	ctx, cancel := context.WithTimeout(ctx, c.RequestTimeout)
	defer cancel()

	req, err := http.NewRequestWithContext(ctx, http.MethodGet, url, nil)
	if err != nil {
		return nil, err
	}

	query := req.URL.Query()

	for key, values := range params {
		query[key] = values // replaces original values
	}

	if bs, err := httputil.DumpRequestOut(req, true); err == nil {
		log.WithContext(ctx).Debugf(">>> %s", string(bs))
	}

	query.Set("api_key", c.Key)

	req.URL.RawQuery = query.Encode()

	resp, err := c.Client.Do(req)

	if bs, err := httputil.DumpResponse(resp, true); err == nil {
		log.WithContext(ctx).Debugf("<<< %s", string(bs))
	}

	if err != nil {
		return nil, err
	}

	defer func() {
		_ = resp.Body.Close()
	}()

	switch resp.StatusCode {
	default:
		return nil, fmt.Errorf(
			"%w: unexpected response status: %d %s",
			grpcerr.ProviderError,
			resp.StatusCode,
			resp.Status,
		)

		// All codes above indicate conventional data exchange,
		// see https://ideal-postcodes.co.uk/documentation/response-codes
	case http.StatusOK,
		http.StatusBadRequest,
		http.StatusUnauthorized,
		http.StatusPaymentRequired,
		http.StatusNotFound,
		http.StatusInternalServerError:
	}

	body, err := io.ReadAll(resp.Body)
	if err != nil {
		return nil, err
	}

	return body, nil
}

// apiError verifies the code from IdealPostcode response
// and returns error with specified message if it is not equal 2000.
func apiError(code float64, message string) error {
	if code != apiStatusOK {
		return fmt.Errorf("%w: provider return error %.0f %s", grpcerr.ProviderError, code, message)
	}

	return nil
}
